@@ -32,13 +32,19 @@ from typing import (
     Generator,
 )
 
+import graphviz
 import ijson
 from grammar_graph import gg
 from graphviz import Digraph
 
-from isla.helpers import is_nonterminal, traverse, TRAVERSE_POSTORDER
+from isla.helpers import (
+    is_nonterminal,
+    traverse,
+    TRAVERSE_POSTORDER,
+    dict_of_lists_to_list_of_dicts,
+)
 from isla.trie import SubtreesTrie
-from isla.type_defs import Path, ParseTree
+from isla.type_defs import Path, ParseTree, CanonicalGrammar
 
 
 class DerivationTree:
@@ -605,6 +611,54 @@ class DerivationTree:
         #     for path in common_paths
         # )
 
+    def expand_one_step(
+        self, canonical_grammar: CanonicalGrammar
+    ) -> List["DerivationTree"]:
+        """
+        Expands this tree one step. This may result in multiple expanded trees
+        if the grammar contains multiple alternatives for the symbol of at least one
+        leaf node.
+
+        :param canonical_grammar: The grammar to use for expansion, in canonical form.
+        :return: All single-step expansions.
+        """
+        result = []
+        nonterminal_expansions: Dict[Path, List[List[DerivationTree]]] = {
+            leaf_path: [
+                [
+                    DerivationTree(child, None if is_nonterminal(child) else [])
+                    for child in expansion
+                ]
+                for expansion in canonical_grammar[leaf_node.value]
+            ]
+            for leaf_path, leaf_node in self.open_leaves()
+        }
+
+        if not nonterminal_expansions:
+            return []
+
+        possible_expansions: List[
+            Dict[Path, List[DerivationTree]]
+        ] = dict_of_lists_to_list_of_dicts(nonterminal_expansions)
+
+        assert (
+            len(possible_expansions) > 1
+            or len(possible_expansions) == 1
+            and possible_expansions[0]
+        )
+
+        for possible_expansion in possible_expansions:
+            new_expanded_tree = self
+            for path, new_children in possible_expansion.items():
+                leaf_node = new_expanded_tree.get_subtree(path)
+                new_expanded_tree = new_expanded_tree.replace_path(
+                    path,
+                    DerivationTree(leaf_node.value, new_children, leaf_node.id),
+                )
+            result.append(new_expanded_tree)
+
+        return result
+
     @staticmethod
     def from_parse_tree(tree: ParseTree):
         result_stack: List[DerivationTree] = []
@@ -814,8 +868,9 @@ class DerivationTree:
     def to_dot(self) -> str:
         dot = Digraph(comment="Derivation Tree")
         dot.attr("node", shape="plain")
+        levels: Dict[int, List[DerivationTree]] = {}
 
-        def action(_, t: DerivationTree):
+        def action(path: Path, t: DerivationTree):
             dot.node(
                 repr(t.id),
                 "<" + html.escape(t.value) + f' <FONT COLOR="gray">({t.id})</FONT>>',
@@ -823,6 +878,24 @@ class DerivationTree:
             for child in t.children or []:
                 dot.edge(repr(t.id), repr(child.id))
 
+            levels.setdefault(len(path), []).append(t)
+
         self.traverse(action)
 
+        # Ensure that nodes at the same levels actually appear at the same levels,
+        # and in the right orders.
+        for nodes in levels.values():
+            if len(nodes) < 2:
+                continue
+
+            child_graph = graphviz.Digraph()
+            child_graph.attr(rank="same")
+
+            for idx in range(len(nodes) - 1):
+                child_graph.edge(
+                    repr(nodes[idx].id), repr(nodes[idx + 1].id), style="invis"
+                )
+
+            dot.subgraph(child_graph)
+
         return dot.source
